import numpy as np
from abc import ABC, abstractmethod


class LeafFunction(ABC):
    """
    Base class leaf value assignment functions
    """
    def __init__(self) -> None:
        self.name = self.__class__.__name__

    def __call__(self, targets) -> float:
        """
        calls '_leafFunc', makes working with this class easier
        """
        return self._leafFunc(targets)

    @abstractmethod
    def _leafFunc(self, targets) -> float:
        """
        needs to be implemented with every daughter class
        """
        pass


class Mode(LeafFunction):
    """
    this leaf function is used for classification tasks
    """
    def __init__(self) -> None:
        super().__init__()

    def _leafFunc(self, targets) -> int:
        if len(targets.shape) > 1 and targets.shape[1] > 1:
            # Compute the mode of class indices using NumPy
            unique, counts = np.unique(targets, return_counts=True, axis=0)
            maxCount = np.max(counts)
            modeValue = unique[counts == maxCount]

            return modeValue[0]

        values, counts = np.unique(targets, return_counts=True)
        return values[counts.argmax()]


class Mean(LeafFunction):
    """
    this leaf function is used for regression tasks
    """
    def __init__(self) -> None:
        super().__init__()

    def _leafFunc(self, targets) -> float:
        return np.mean(targets, axis=0)


class Median(LeafFunction):
    """
    this leaf function is used for regression tasks
    """
    def __init__(self) -> None:
        super().__init__()

    def _leafFunc(self, targets) -> float:
        return np.median(targets, axis=0)


class Probabilities(LeafFunction):
    """
    This leaf function is used for classification tasks where you want
    probabilities for each class instead of predicting a single class.
    """
    def __init__(self) -> None:
        super().__init__()

    def _leafFunc(self, targets) -> dict:
        unique, counts = np.unique(targets, return_counts=True)
        probs = counts / len(targets)
        return dict(zip(unique, probs))


class Confidence(LeafFunction):
    """
    This leaf function is used for classification tasks where you want
    the confidence of the mode class.
    """
    def __init__(self) -> None:
        super().__init__()

    def _leafFunc(self, targets) -> float:
        values, counts = np.unique(targets, return_counts=True)
        mode_count = counts.max()
        confidence = mode_count / len(targets)
        return confidence
